"""Small library for in-memory cardinality computing."""

VERSION = (0, 1, 0)

__version__ = '.'.join(map(str, VERSION[0:3]))
__author__ = 'Lipin Dmitriy'
__contact__ = 'blackwithwhite666@gmail.com'
__homepage__ = 'https://github.com/blackwithwhite666/pyhll'
__docformat__ = 'restructuredtext'

# -eof meta-

from .hll import (
    Cardinality,
    precision_for_error,
    error_for_precision,
    bytes_for_precision,
    MIN_PRECISION,
    MAX_PRECISION,
)





"""This is a wrapper to HyperLogLog implementation on C.

HyperLogLog's are a relatively new sketching data structure.
They are used to estimate cardinality, i.e. the unique number
of items in a set. They are based on the observation that any bit
in a "good" hash function is indepedenent of any other bit
and that the probability of getting a string of N bits all set
to the same value is 1/(2^N). There is a lot more in the math,
but that is the basic intuition. What is even more incredible
is that the storage required to do the counting is log(log(N)).
So with a 6 bit register, we can count well into the trillions.
"""
from libc.stdlib cimport malloc, free


cdef extern from "stdint.h" nogil:

    ctypedef unsigned int uint32_t
    ctypedef unsigned int uint64_t


cdef extern from "hll_constants.h":
    pass


cdef extern from "hll.h":

    int HLL_MIN_PRECISION
    int HLL_MAX_PRECISION

    ctypedef struct hll_t:
        unsigned char precision
        uint32_t *registers

    int hll_init(unsigned char precision, hll_t *h)
    int hll_destroy(hll_t *h)

    void hll_add(hll_t *h, char *key)
    void hll_add_hash(hll_t *h, uint64_t hash)
    double hll_size(hll_t *h)
    int hll_precision_for_error(double err)
    double hll_error_for_precision(int prec)
    uint64_t hll_bytes_for_precision(int prec)


MIN_PRECISION = HLL_MIN_PRECISION
MAX_PRECISION = HLL_MAX_PRECISION


cdef class Cardinality(object):
    """Compute cardinality of some set using HyperLogLog algorithm.

    :param prec: precision to use,
    :type prec: int
    """

    cdef hll_t *_c_hll
    cdef int _initialized
    cdef int _precision

    def __cinit__(self, int prec=12):
        self._initialized = -1
        if not (HLL_MIN_PRECISION <= prec <= HLL_MAX_PRECISION):
            raise ValueError('Wrong precision value "{0!r}" given'.format(prec))
        self._precision = prec
        self._c_hll = <hll_t *>malloc(sizeof(hll_t))
        if self._c_hll is NULL:
            raise MemoryError("Can't create hll_t struct")
        self._initialized = hll_init(self._precision, self._c_hll)
        assert self._initialized == 0, "initialization failed"

    def __dealloc__(self):
        if self._c_hll is not NULL:
            if self._initialized == 0:
                hll_destroy(self._c_hll)
            free(self._c_hll)

    property precision:

        def __get__(self):
            return self._precision

    property size:

        def __get__(self):
            return hll_size(self._c_hll)

    property count:

        def __get__(self):
            return int(self.size)

    def __len__(self):
        return self.count

    def add(self, bytes s):
        """Add one string."""
        cdef char* c_string = s
        hll_add(self._c_hll, c_string)

    def update(self, strings):
        """Add multiple strings at once."""
        for s in strings:
            self.add(s)

    def __repr__(self):
        return ('<{0}(size={2.size}) at {1}>'.
                format(self.__class__.__name__, hex(id(self)), self))


def precision_for_error(double err):
    """Get precision for given error value."""
    cdef int result = hll_precision_for_error(err)
    if result == -1:
        raise ValueError('Wrong error value "{0!r}" given'.format(err))
    return result


def error_for_precision(int prec):
    """Return error for given precision."""
    cdef double result = hll_error_for_precision(prec)
    if result == 0:
        raise ValueError('Wrong precision value "{0!r}" given'.format(prec))
    return result


def bytes_for_precision(int prec):
    """Get needed memory in bytes for given precision."""
    cdef uint64_t result = hll_bytes_for_precision(prec)
    if result == 0:
        raise ValueError('Wrong precision value "{0!r}" given'.format(prec))
    return result



from __future__ import absolute_import

from pyhll.tests.base import TestCase
from pyhll import Cardinality, MIN_PRECISION, MAX_PRECISION


class CardinalityTest(TestCase):

    def test_construct(self):
        with self.assertRaises(ValueError):
            Cardinality(MIN_PRECISION - 1)
        with self.assertRaises(ValueError):
            Cardinality(MAX_PRECISION + 1)
        self.assertEqual(MIN_PRECISION, Cardinality(MIN_PRECISION).precision)
        self.assertEqual(MAX_PRECISION, Cardinality(MAX_PRECISION).precision)

    def test_size(self):
        t = (b'foo', b'bar')
        c = Cardinality()
        c.update(t)
        self.assertAlmostEqual(len(t), c.size, places=1)

    def test_count(self):
        t = (b'foo', b'bar')
        c = Cardinality()
        c.update((b'foo', b'bar'))
        self.assertEqual(len(t), c.count)
        self.assertEqual(len(t), len(c))

    def test_add(self):
        c = Cardinality()
        c.add(b'foo')
        self.assertEqual(1, len(c))
        c.add(b'bar')
        self.assertEqual(2, len(c))
        with self.assertRaises(TypeError):
            c.add(u'foo')
        with self.assertRaises(TypeError):
            c.add(1)

    def test_update(self):
        c = Cardinality()
        c.update([b'foo', b'baz'])
        self.assertEqual(2, len(c))
        c.update([b'bar', b'boz'])
        self.assertEqual(4, len(c))



from __future__ import absolute_import

from pyhll.tests.base import TestCase
from pyhll import (
    bytes_for_precision,
    error_for_precision,
    precision_for_error,
    MIN_PRECISION,
    MAX_PRECISION,
)


class UtilsTest(TestCase):

    def test_bytes_for_precision(self):
        with self.assertRaises(ValueError):
            bytes_for_precision(MIN_PRECISION - 1)
        with self.assertRaises(ValueError):
            bytes_for_precision(MAX_PRECISION + 1)
        self.assertGreater(bytes_for_precision(MIN_PRECISION), 0)
        self.assertGreater(bytes_for_precision(MAX_PRECISION), 0)
        self.assertEqual(16, bytes_for_precision(4))
        self.assertEqual(3280, bytes_for_precision(12))
        self.assertEqual(52432, bytes_for_precision(16))

    def test_error_for_precision(self):
        with self.assertRaises(ValueError):
            error_for_precision(MIN_PRECISION - 1)
        with self.assertRaises(ValueError):
            error_for_precision(MAX_PRECISION + 1)
        self.assertGreater(error_for_precision(MIN_PRECISION), 0)
        self.assertGreater(error_for_precision(MAX_PRECISION), 0)
        self.assertAlmostEqual(0.26, error_for_precision(4))
        self.assertAlmostEqual(0.01625, error_for_precision(12))
        self.assertAlmostEqual(0.0040625, error_for_precision(16))

    def test_precision_for_error(self):
        with self.assertRaises(ValueError):
            precision_for_error(2.0)
        with self.assertRaises(ValueError):
            precision_for_error(-2.0)
        self.assertEqual(4, precision_for_error(0.26))
        self.assertEqual(12, precision_for_error(0.01625))
        self.assertEqual(16, precision_for_error(0.0040625))



import os
import re
import sys
import errno
import subprocess

from setuptools import setup, Extension, find_packages
from setuptools.command.sdist import sdist
from setuptools.command.build_ext import build_ext

#-----------------------------------------------------------------------------
# Flags and default values.
#-----------------------------------------------------------------------------

PACKAGE = 'pyhll'

cmdclass = {}
extensions = []
extension_kwargs = {}


# try to find cython
try:
    from Cython.Distutils import build_ext as build_ext_c
    cython_installed = True
except ImportError:
    cython_installed = False


# current location
here = os.path.abspath(os.path.dirname(__file__))


def exec_process(cmdline, silent=True, input=None, **kwargs):
    """Execute a subprocess and returns the returncode, stdout buffer and stderr buffer.
    Optionally prints stdout and stderr while running."""
    try:
        sub = subprocess.Popen(args=cmdline, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, **kwargs)
        stdout, stderr = sub.communicate(input=input)
        returncode = sub.returncode
        if not silent:
            sys.stdout.write(stdout)
            sys.stderr.write(stderr)
    except OSError as e:
        if e.errno == errno.ENOENT:
            raise RuntimeError('"%s" is not present on this system' % cmdline[0])
        else:
            raise
    if returncode != 0:
        raise RuntimeError('Got return value %d while executing "%s", stderr output was:\n%s' % (returncode, " ".join(cmdline), stderr.decode("utf-8").rstrip("\n")))
    return stdout


#-----------------------------------------------------------------------------
# Commands
#-----------------------------------------------------------------------------

class CheckSDist(sdist):
    """Custom sdist that ensures Cython has compiled all pyx files to c."""

    def initialize_options(self):
        sdist.initialize_options(self)
        self._pyxfiles = []
        for root, dirs, files in os.walk(PACKAGE):
            for f in files:
                if f.endswith('.pyx'):
                    self._pyxfiles.append(os.path.join(root, f))

    def run(self):
        if 'cython' in cmdclass:
            self.run_command('cython')
        else:
            for pyxfile in self._pyxfiles:
                cfile = pyxfile[:-3] + 'c'
                msg = "C-source file '%s' not found."%(cfile)+\
                " Run 'setup.py cython' before sdist."
                assert os.path.isfile(cfile), msg
        sdist.run(self)

cmdclass['sdist'] = CheckSDist


class BaseMurmurMixin:

    murmur_dir = os.path.join(here, 'deps', 'murmurhash')

    def build_murmur(self):
        cflags = '-fPIC'
        env = os.environ.copy()
        env['CPPFLAGS'] = ' '.join(x for x in (cflags, env.get('CPPFLAGS', None)) if x)
        exec_process(['sh', 'autogen.sh'], cwd=self.murmur_dir, env=env, silent=True)
        exec_process(['./configure'], cwd=self.murmur_dir, env=env, silent=True)
        exec_process(['make'], cwd=self.murmur_dir, env=env, silent=False)

    def prepare_extensions(self):
        self.murmur_lib = os.path.join(self.murmur_dir, '.libs', 'libMurmurHash3.a')
        if not os.path.exists(os.path.join(self.murmur_dir, '.libs')):
            self.build_murmur()
        self.extensions[0].extra_objects.extend([self.murmur_lib])
        self.compiler.add_include_dir(self.murmur_lib)


if cython_installed:

    class CythonCommand(build_ext_c):
        """Custom distutils command subclassed from Cython.Distutils.build_ext
        to compile pyx->c, and stop there. All this does is override the
        C-compile method build_extension() with a no-op."""

        description = "Compile Cython sources to C"

        def build_extension(self, ext):
            pass


    class MurmurMixin(BaseMurmurMixin):

        def build_extensions(self):
            self.prepare_extensions()
            build_ext_c.build_extensions(self)


    class zbuild_ext(MurmurMixin, build_ext_c):

        def run(self):
            from distutils.command.build_ext import build_ext as _build_ext
            return _build_ext.run(self)

    cmdclass['cython'] = CythonCommand
    cmdclass['build_ext'] = zbuild_ext

else:

    class CheckingBuildExt(BaseMurmurMixin, build_ext):
        """Subclass build_ext to get clearer report if Cython is neccessary."""

        def check_cython_extensions(self, extensions):
            for ext in extensions:
                for src in ext.sources:
                    msg = "Cython-generated file '%s' not found." % src
                    assert os.path.exists(src), msg

        def build_extensions(self):
            self.prepare_extensions()
            self.check_cython_extensions(self.extensions)
            self.check_extensions_list(self.extensions)

            for ext in self.extensions:
                self.build_extension(ext)

    cmdclass['build_ext'] = CheckingBuildExt


#-----------------------------------------------------------------------------
# Extensions
#-----------------------------------------------------------------------------


suffix = '.pyx' if cython_installed else '.c'


def source_extension(name):
    parts = name.split('.')
    parts[-1] = parts[-1] + suffix
    return os.path.join(PACKAGE, *parts)


def prepare_sources(sources):
    def to_string(s):
        if isinstance(s, unicode):
            return s.encode('utf-8')
        return s
    return [to_string(source) for source in sources]


modules = {
    'hll': dict(
        include_dirs=[os.path.join(here, 'src')],
        sources=[
            source_extension('hll'),
            os.path.join(here, 'src', 'hll.c'),
            os.path.join(here, 'src', 'hll_constants.c'),
        ],
        extra_compile_args=['-std=c99'],
    ),
}

# collect extensions
for module, kwargs in modules.items():
    kwargs = dict(extension_kwargs, **kwargs)
    kwargs.setdefault('sources', [source_extension(module)])
    kwargs['sources'] = prepare_sources(kwargs['sources'])
    ext = Extension('{0}.{1}'.format(PACKAGE, module), **kwargs)
    if suffix == '.pyx' and ext.sources[0].endswith('.c'):
        # undo setuptools stupidly clobbering cython sources:
        ext.sources = kwargs['sources']
    extensions.append(ext)


#-----------------------------------------------------------------------------
# Description, version and other meta information.
#-----------------------------------------------------------------------------

re_meta = re.compile(r'__(\w+?)__\s*=\s*(.*)')
re_vers = re.compile(r'VERSION\s*=\s*\((.*?)\)')
re_doc = re.compile(r'^"""(.+?)"""')
rq = lambda s: s.strip("\"'")


def add_default(m):
    attr_name, attr_value = m.groups()
    return ((attr_name, rq(attr_value)),)


def add_version(m):
    v = list(map(rq, m.groups()[0].split(', ')))
    return (('VERSION', '.'.join(v[0:3]) + ''.join(v[3:])),)


def add_doc(m):
    return (('doc', m.groups()[0]),)

pats = {re_meta: add_default,
        re_vers: add_version,
        re_doc: add_doc}
meta_fh = open(os.path.join(here, '{0}/__init__.py'.format(PACKAGE)))
try:
    meta = {}
    for line in meta_fh:
        if line.strip() == '# -eof meta-':
            break
        for pattern, handler in pats.items():
            m = pattern.match(line.strip())
            if m:
                meta.update(handler(m))
finally:
    meta_fh.close()

with open(os.path.join(here, 'README.rst')) as f:
    README = f.read()

with open(os.path.join(here, 'CHANGES.rst')) as f:
    CHANGES = f.read()


#-----------------------------------------------------------------------------
# Setup
#-----------------------------------------------------------------------------

setup(
    name=PACKAGE,
    version=meta['VERSION'],
    description=meta['doc'],
    author=meta['author'],
    author_email=meta['contact'],
    url=meta['homepage'],
    long_description=README + '\n\n' + CHANGES,
    keywords='thrift soa',
    license='BSD',
    cmdclass=cmdclass,
    ext_modules=extensions,
    packages=find_packages(),
    install_requires=[],
    zip_safe=False,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 2.7",
    ],
)
